@@ -47,20 +47,27 @@ pub const MAX_DGRAM_OVERHEAD: usize = 2;
 pub const MAX_STREAM_OVERHEAD: usize = 12;
 pub const MAX_STREAM_SIZE: u64 = 1 << 62;
 
-#[derive(Clone, Debug, PartialEq)]
+#[derive(Clone, Debug, PartialEq, Eq)]
 pub struct EcnCounts {
     ect0_count: u64,
     ect1_count: u64,
     ecn_ce_count: u64,
 }
 
-#[derive(Clone, PartialEq)]
+#[derive(Clone, PartialEq, Eq)]
 pub enum Frame {
     Padding {
         len: usize,
     },
 
-    Ping,
+    Ping {
+        // Attach metadata to the Ping frame. This doesn't appear on the wire,
+        // but does tell us if this frame was part of a PMTUD probe and how
+        // large the probe was. This will only show up on sent frames and be
+        // None otherwise. This is the total size of the QUIC packet in the
+        // probe.
+        mtu_probe: Option<usize>,
+    },
 
     ACK {
         ack_delay: u64,
@@ -185,8 +192,6 @@ impl Frame {
     ) -> Result<Frame> {
         let frame_type = b.get_varint()?;
 
-        // println!("GOT FRAME {:x}", frame_type);
-
         let frame = match frame_type {
             0x00 => {
                 let mut len = 1;
@@ -200,7 +205,7 @@ impl Frame {
                 Frame::Padding { len }
             },
 
-            0x01 => Frame::Ping,
+            0x01 => Frame::Ping { mtu_probe: None },
 
             0x02..=0x03 => parse_ack_frame(frame_type, b)?,
 
@@ -223,8 +228,15 @@ impl Frame {
                 Frame::Crypto { data }
             },
 
-            0x07 => Frame::NewToken {
-                token: b.get_bytes_with_varint_length()?.to_vec(),
+            0x07 => {
+                let len = b.get_varint()?;
+                if len == 0 {
+                    return Err(Error::InvalidFrame);
+                }
+
+                Frame::NewToken {
+                    token: b.get_bytes(len as usize)?.to_vec(),
+                }
             },
 
             0x08..=0x0f => parse_stream_frame(frame_type, b)?,
@@ -263,15 +275,25 @@ impl Frame {
                 limit: b.get_varint()?,
             },
 
-            0x18 => Frame::NewConnectionId {
-                seq_num: b.get_varint()?,
-                retire_prior_to: b.get_varint()?,
-                conn_id: b.get_bytes_with_u8_length()?.to_vec(),
-                reset_token: b
-                    .get_bytes(16)?
-                    .buf()
-                    .try_into()
-                    .map_err(|_| Error::BufferTooShort)?,
+            0x18 => {
+                let seq_num = b.get_varint()?;
+                let retire_prior_to = b.get_varint()?;
+                let conn_id_len = b.get_u8()?;
+
+                if !(1..=packet::MAX_CID_LEN).contains(&conn_id_len) {
+                    return Err(Error::InvalidFrame);
+                }
+
+                Frame::NewConnectionId {
+                    seq_num,
+                    retire_prior_to,
+                    conn_id: b.get_bytes(conn_id_len as usize)?.to_vec(),
+                    reset_token: b
+                        .get_bytes(16)?
+                        .buf()
+                        .try_into()
+                        .map_err(|_| Error::BufferTooShort)?,
+                }
             },
 
             0x19 => Frame::RetireConnectionId {
@@ -361,7 +383,7 @@ impl Frame {
                 }
             },
 
-            Frame::Ping => {
+            Frame::Ping { .. } => {
                 b.put_varint(0x01)?;
             },
 
@@ -428,7 +450,7 @@ impl Frame {
             },
 
             Frame::Crypto { data } => {
-                encode_crypto_header(data.off() as u64, data.len() as u64, b)?;
+                encode_crypto_header(data.off(), data.len() as u64, b)?;
 
                 b.put_bytes(data)?;
             },
@@ -445,7 +467,7 @@ impl Frame {
             Frame::Stream { stream_id, data } => {
                 encode_stream_header(
                     *stream_id,
-                    data.off() as u64,
+                    data.off(),
                     data.len() as u64,
                     data.fin(),
                     b,
@@ -580,7 +602,7 @@ impl Frame {
         match self {
             Frame::Padding { len } => *len,
 
-            Frame::Ping => 1,
+            Frame::Ping { .. } => 1,
 
             Frame::ACK {
                 ack_delay,
@@ -641,7 +663,7 @@ impl Frame {
 
             Frame::Crypto { data } => {
                 1 + // frame type
-                octets::varint_len(data.off() as u64) + // offset
+                octets::varint_len(data.off()) + // offset
                 2 + // length, always encode as 2-byte varint
                 data.len() // data
             },
@@ -662,7 +684,7 @@ impl Frame {
             Frame::Stream { stream_id, data } => {
                 1 + // frame type
                 octets::varint_len(*stream_id) + // stream_id
-                octets::varint_len(data.off() as u64) + // offset
+                octets::varint_len(data.off()) + // offset
                 2 + // length, always encode as 2-byte varint
                 data.len() // data
             },
@@ -800,12 +822,28 @@ impl Frame {
         )
     }
 
+    pub fn probing(&self) -> bool {
+        matches!(
+            self,
+            Frame::Padding { .. } |
+                Frame::NewConnectionId { .. } |
+                Frame::PathChallenge { .. } |
+                Frame::PathResponse { .. }
+        )
+    }
+
     #[cfg(feature = "qlog")]
     pub fn to_qlog(&self) -> QuicFrame {
         match self {
-            Frame::Padding { .. } => QuicFrame::Padding,
+            Frame::Padding { len } => QuicFrame::Padding {
+                length: None,
+                payload_length: *len as u32,
+            },
 
-            Frame::Ping { .. } => QuicFrame::Ping,
+            Frame::Ping { .. } => QuicFrame::Ping {
+                length: None,
+                payload_length: None,
+            },
 
             Frame::ACK {
                 ack_delay,
@@ -832,6 +870,8 @@ impl Frame {
                     ect1,
                     ect0,
                     ce,
+                    length: None,
+                    payload_length: None,
                 }
             },
 
@@ -843,6 +883,8 @@ impl Frame {
                 stream_id: *stream_id,
                 error_code: *error_code,
                 final_size: *final_size,
+                length: None,
+                payload_length: None,
             },
 
             Frame::StopSending {
@@ -851,6 +893,8 @@ impl Frame {
             } => QuicFrame::StopSending {
                 stream_id: *stream_id,
                 error_code: *error_code,
+                length: None,
+                payload_length: None,
             },
 
             Frame::Crypto { data } => QuicFrame::Crypto {
@@ -864,15 +908,23 @@ impl Frame {
             },
 
             Frame::NewToken { token } => QuicFrame::NewToken {
-                length: token.len().to_string(),
-                token: "TODO: update to qlog-02 token format".to_string(),
+                token: qlog::Token {
+                    // TODO: pick the token type some how
+                    ty: Some(qlog::TokenType::Retry),
+                    raw: Some(qlog::events::RawInfo {
+                        data: qlog::HexSlice::maybe_string(Some(token)),
+                        length: Some(token.len() as u64),
+                        payload_length: None,
+                    }),
+                    details: None,
+                },
             },
 
             Frame::Stream { stream_id, data } => QuicFrame::Stream {
                 stream_id: *stream_id,
-                offset: data.off() as u64,
+                offset: data.off(),
                 length: data.len() as u64,
-                fin: data.fin().then(|| true),
+                fin: data.fin().then_some(true),
                 raw: None,
             },
 
@@ -929,13 +981,15 @@ impl Frame {
                 seq_num,
                 retire_prior_to,
                 conn_id,
-                ..
+                reset_token,
             } => QuicFrame::NewConnectionId {
                 sequence_number: *seq_num as u32,
                 retire_prior_to: *retire_prior_to as u32,
-                length: conn_id.len() as u64,
-                connection_id: "TODO: update to qlog-02 token format".to_string(),
-                reset_token: "TODO: update to qlog-02 token format".to_string(),
+                connection_id_length: Some(conn_id.len() as u8),
+                connection_id: format!("{}", qlog::HexSlice::new(conn_id)),
+                stateless_reset_token: qlog::HexSlice::maybe_string(Some(
+                    reset_token,
+                )),
             },
 
             Frame::RetireConnectionId { seq_num } =>
@@ -951,19 +1005,19 @@ impl Frame {
             Frame::ConnectionClose {
                 error_code, reason, ..
             } => QuicFrame::ConnectionClose {
-                error_space: ErrorSpace::TransportError,
-                error_code: *error_code,
-                raw_error_code: None, // raw error is no different for us
-                reason: Some(String::from_utf8(reason.clone()).unwrap()),
+                error_space: Some(ErrorSpace::TransportError),
+                error_code: Some(*error_code),
+                error_code_value: None, // raw error is no different for us
+                reason: Some(String::from_utf8_lossy(reason).into_owned()),
                 trigger_frame_type: None, // don't know trigger type
             },
 
             Frame::ApplicationClose { error_code, reason } =>
                 QuicFrame::ConnectionClose {
-                    error_space: ErrorSpace::ApplicationError,
-                    error_code: *error_code,
-                    raw_error_code: None, // raw error is no different for us
-                    reason: Some(String::from_utf8(reason.clone()).unwrap()),
+                    error_space: Some(ErrorSpace::ApplicationError),
+                    error_code: Some(*error_code),
+                    error_code_value: None, // raw error is no different for us
+                    reason: Some(String::from_utf8_lossy(reason).into_owned()),
                     trigger_frame_type: None, // don't know trigger type
                 },
 
@@ -986,11 +1040,11 @@ impl std::fmt::Debug for Frame {
     fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
         match self {
             Frame::Padding { len } => {
-                write!(f, "PADDING len={}", len)?;
+                write!(f, "PADDING len={len}")?;
             },
 
-            Frame::Ping => {
-                write!(f, "PING")?;
+            Frame::Ping { mtu_probe } => {
+                write!(f, "PING mtu_probe={mtu_probe:?}")?;
             },
 
             Frame::ACK {
@@ -1000,8 +1054,7 @@ impl std::fmt::Debug for Frame {
             } => {
                 write!(
                     f,
-                    "ACK delay={} blocks={:?} ecn_counts={:?}",
-                    ack_delay, ranges, ecn_counts
+                    "ACK delay={ack_delay} blocks={ranges:?} ecn_counts={ecn_counts:?}"
                 )?;
             },
 
@@ -1012,8 +1065,7 @@ impl std::fmt::Debug for Frame {
             } => {
                 write!(
                     f,
-                    "RESET_STREAM stream={} err={:x} size={}",
-                    stream_id, error_code, final_size
+                    "RESET_STREAM stream={stream_id} err={error_code:x} size={final_size}"
                 )?;
             },
 
@@ -1021,11 +1073,7 @@ impl std::fmt::Debug for Frame {
                 stream_id,
                 error_code,
             } => {
-                write!(
-                    f,
-                    "STOP_SENDING stream={} err={:x}",
-                    stream_id, error_code
-                )?;
+                write!(f, "STOP_SENDING stream={stream_id} err={error_code:x}")?;
             },
 
             Frame::Crypto { data } => {
@@ -1033,11 +1081,11 @@ impl std::fmt::Debug for Frame {
             },
 
             Frame::CryptoHeader { offset, length } => {
-                write!(f, "CRYPTO off={} len={}", offset, length)?;
+                write!(f, "CRYPTO off={offset} len={length}")?;
             },
 
-            Frame::NewToken { .. } => {
-                write!(f, "NEW_TOKEN (TODO)")?;
+            Frame::NewToken { token } => {
+                write!(f, "NEW_TOKEN len={}", token.len())?;
             },
 
             Frame::Stream { stream_id, data } => {
@@ -1059,61 +1107,67 @@ impl std::fmt::Debug for Frame {
             } => {
                 write!(
                     f,
-                    "STREAM id={} off={} len={} fin={}",
-                    stream_id, offset, length, fin
+                    "STREAM id={stream_id} off={offset} len={length} fin={fin}"
                 )?;
             },
 
             Frame::MaxData { max } => {
-                write!(f, "MAX_DATA max={}", max)?;
+                write!(f, "MAX_DATA max={max}")?;
             },
 
             Frame::MaxStreamData { stream_id, max } => {
-                write!(f, "MAX_STREAM_DATA stream={} max={}", stream_id, max)?;
+                write!(f, "MAX_STREAM_DATA stream={stream_id} max={max}")?;
             },
 
             Frame::MaxStreamsBidi { max } => {
-                write!(f, "MAX_STREAMS type=bidi max={}", max)?;
+                write!(f, "MAX_STREAMS type=bidi max={max}")?;
             },
 
             Frame::MaxStreamsUni { max } => {
-                write!(f, "MAX_STREAMS type=uni max={}", max)?;
+                write!(f, "MAX_STREAMS type=uni max={max}")?;
             },
 
             Frame::DataBlocked { limit } => {
-                write!(f, "DATA_BLOCKED limit={}", limit)?;
+                write!(f, "DATA_BLOCKED limit={limit}")?;
             },
 
             Frame::StreamDataBlocked { stream_id, limit } => {
                 write!(
                     f,
-                    "STREAM_DATA_BLOCKED stream={} limit={}",
-                    stream_id, limit
+                    "STREAM_DATA_BLOCKED stream={stream_id} limit={limit}"
                 )?;
             },
 
             Frame::StreamsBlockedBidi { limit } => {
-                write!(f, "STREAMS_BLOCKED type=bidi limit={}", limit)?;
+                write!(f, "STREAMS_BLOCKED type=bidi limit={limit}")?;
             },
 
             Frame::StreamsBlockedUni { limit } => {
-                write!(f, "STREAMS_BLOCKED type=uni limit={}", limit)?;
+                write!(f, "STREAMS_BLOCKED type=uni limit={limit}")?;
             },
 
-            Frame::NewConnectionId { .. } => {
-                write!(f, "NEW_CONNECTION_ID (TODO)")?;
+            Frame::NewConnectionId {
+                seq_num,
+                retire_prior_to,
+                conn_id,
+                reset_token,
+            } => {
+                write!(
+                    f,
+                    "NEW_CONNECTION_ID seq_num={seq_num} retire_prior_to={retire_prior_to} conn_id={conn_id:02x?} reset_token={reset_token:02x?}",
+                )?;
             },
 
-            Frame::RetireConnectionId { .. } => {
-                write!(f, "RETIRE_CONNECTION_ID (TODO)")?;
+            Frame::RetireConnectionId { seq_num } => {
+                write!(f, "RETIRE_CONNECTION_ID seq_num={seq_num}")?;
             },
 
             Frame::PathChallenge { data } => {
-                write!(f, "PATH_CHALLENGE data={:02x?}", data)?;
+                write!(f, "PATH_CHALLENGE data={data:02x?}")?;
             },
 
             Frame::PathResponse { data } => {
-                write!(f, "PATH_RESPONSE data={:02x?}", data)?;
+                write!(f, "PATH_RESPONSE data={data:02x?}")?;
             },
 
             Frame::ConnectionClose {
@@ -1123,16 +1177,14 @@ impl std::fmt::Debug for Frame {
             } => {
                 write!(
                     f,
-                    "CONNECTION_CLOSE err={:x} frame={:x} reason={:x?}",
-                    error_code, frame_type, reason
+                    "CONNECTION_CLOSE err={error_code:x} frame={frame_type:x} reason={reason:x?}"
                 )?;
             },
 
             Frame::ApplicationClose { error_code, reason } => {
                 write!(
                     f,
-                    "APPLICATION_CLOSE err={:x} reason={:x?}",
-                    error_code, reason
+                    "APPLICATION_CLOSE err={error_code:x} reason={reason:x?}"
                 )?;
             },
 
@@ -1145,7 +1197,7 @@ impl std::fmt::Debug for Frame {
             },
 
             Frame::DatagramHeader { length } => {
-                write!(f, "DATAGRAM len={}", length)?;
+                write!(f, "DATAGRAM len={length}")?;
             },
         }
 
@@ -1342,7 +1394,7 @@ mod tests {
     fn ping() {
         let mut d = [42; 128];
 
-        let frame = Frame::Ping;
+        let frame = Frame::Ping { mtu_probe: None };
 
         let wire_len = {
             let mut b = octets::OctetsMut::with_slice(&mut d);
@@ -1350,7 +1402,7 @@ mod tests {
         };
 
         assert_eq!(wire_len, 1);
-        assert_eq!(&d[..wire_len], [0x01 as u8]);
+        assert_eq!(&d[..wire_len], [0x01_u8]);
 
         let mut b = octets::Octets::with_slice(&d);
         assert_eq!(Frame::from_bytes(&mut b, packet::Type::Short), Ok(frame));
@@ -2043,19 +2095,19 @@ mod tests {
 
         assert_eq!(wire_len, 15);
 
-        let mut b = octets::Octets::with_slice(&mut d);
+        let mut b = octets::Octets::with_slice(&d);
         assert_eq!(
             Frame::from_bytes(&mut b, packet::Type::Short),
             Ok(frame.clone())
         );
 
-        let mut b = octets::Octets::with_slice(&mut d);
+        let mut b = octets::Octets::with_slice(&d);
         assert!(Frame::from_bytes(&mut b, packet::Type::Initial).is_err());
 
-        let mut b = octets::Octets::with_slice(&mut d);
+        let mut b = octets::Octets::with_slice(&d);
         assert!(Frame::from_bytes(&mut b, packet::Type::ZeroRTT).is_ok());
 
-        let mut b = octets::Octets::with_slice(&mut d);
+        let mut b = octets::Octets::with_slice(&d);
         assert!(Frame::from_bytes(&mut b, packet::Type::Handshake).is_err());
 
         let frame_data = match &frame {
